"""

This file initializes the coolest module
and provides some basic information about the package.

"""

# Set the package release version
version_info = (0, 1, 7)
__version__ = '.'.join(str(c) for c in version_info)

# Set the package details
__author__ = 'Aymeric Galan & COOLEST contributors'
__email__ = 'aymeric.galan@gmail.com'
__year__ = '2021-2024'
__credits__ = 'COOLEST contributors'
__url__ = 'https://github.com/aymgal/COOLEST'
__description__ = 'Standard for Strong Gravitational Lensing Analyses'
__license__ = 'GPL-3.0'
